import Line from "./Line.mjs";
import Expr from "./Expr.mjs";

const Invalid = "invalid";

const Program = {
	Invalid: Invalid,
	Valid: "valid",
};

export function parse (code) {
	const lines = code
		.split("\n")
		.filter(l => l.length > 0)
		.map(parse_line);

	if (lines.some(l => l.type === Line.Invalid)) {
		return {
			type: Program.Invalid,
		};
	}

	return {
		type: Program.Valid,
		lines: lines,
	};
}

export function parse_line (line) {
	if (line.length === 0) {
		return {
			type: Line.Empty,
		};
	}

	if (line[0] === "#") {
		return {
			type: Line.Comment,
		};
	}

	if (line[0] === "!") {
		return {
			type: Line.Special,
			value: line,
		};
	}

	const tokens = line.split(/\s+/g).filter(t => t.length > 0);

	if (tokens.length >= 3 && tokens[1] === "=") {
		return parse_binding(tokens);
	}

	return {
		type: Line.Expression,
		value: parse_expression(tokens),
	};
}

function parse_binding (tokens) {
	function invalid () {
		return {
			type: Line.Invalid,
		};
	}

	if (tokens.length < 3) return invalid();

	// TODO: handle parentheses here.

	const name = tokens[0];
	if (!is_valid_name(name)) return invalid();

	const eq = tokens[1];
	if (eq !== "=") return invalid();

	const expr = tokens.slice(2);
	const body = parse_expression(expr);

	if (body.type == Invalid) return invalid();

	return {
		type: Line.Binding,
		name: name,
		body: body,
	};
}

function is_valid_name (name) {
	if (name.length === 0) return false;
	if (name[0] === "#" || name[0] === "!") return false;
	if (name === "=") return false;
	if (name.includes("(") || name.includes(")")) return false;
	if (is_right_arrow(name)) return false;
	return true;
}

const right_arrows = [
	"->",
	"=>",
	">",
	"→",
];
function is_right_arrow (str) {
	return right_arrows.includes(str);
}

function Token (value) {
	return {
		type: Expr.Token,
		value: value,
	};
}

function parse_expression (tokens) {
	if (tokens.length === 0) {
		return {
			type: Expr.Invalid,
		};
	}

	if (tokens.length === 1) {
		return Token(tokens[0]);
	}

	if (tokens.some(is_right_arrow)) {
		return parse_function (tokens);
	}

	return parse_application(tokens);
}

function parse_function (tokens) {
	function invalid () {
		return {
			type: Expr.Invalid,
		};
	}

	const arrow_index = tokens.findIndex(is_right_arrow);
	if (arrow_index === -1) return invalid();

	const args = tokens.slice(0, arrow_index);
	if (args.length === 0) return invalid();
	if (args.some(arg => !is_valid_name(arg))) return invalid();

	const body = parse_expression(tokens.slice(arrow_index + 1));
	if (body.type === Expr.Invalid) return invalid();

	return {
		type: Expr.Fn,
		args: args,
		body: body,
	};
}

function parse_application (tokens) {
	if (tokens.length < 2) {
		return {
			type: Expr.Invalid,
		};
	}

	if (tokens.length === 2) {
		return {
			type: Expr.Application,
			fn: Token(tokens[0]),
			arg: Token(tokens[1]),
		};
	}

	return {
		type: Expr.Application,
		fn: parse_application(tokens.slice(0, 2)),
		arg: parse_expression(tokens.slice(2)),
	};
}
